;***********************************************************************
;* HashTable Library by diablo2oo2                                     *
;* ------------------------------------------------------------------- *
;* -put & get data via key (key can be a string)                       *
;* -save/load the hashtable to a file or memory                        *
;* -hashtable can contain another hashtable                            *
;***********************************************************************

.586p					
.model flat, stdcall
option casemap :none

include				\masm32\include\windows.inc
include				\masm32\macros\macros.asm

include				\masm32\include\user32.inc
include				\masm32\include\kernel32.inc
include				\masm32\include\shell32.inc
include				\masm32\include\advapi32.inc
include				\masm32\include\gdi32.inc
include				\masm32\include\comctl32.inc
include				\masm32\include\comdlg32.inc
include				\masm32\include\winmm.inc
include				\masm32\include\masm32.inc
include 			\masm32\include\msvcrt.inc

includelib			\masm32\lib\user32.lib
includelib			\masm32\lib\kernel32.lib
includelib			\masm32\lib\shell32.lib
includelib			\masm32\lib\advapi32.lib
includelib			\masm32\lib\gdi32.lib
includelib			\masm32\lib\comctl32.lib
includelib			\masm32\lib\comdlg32.lib
includelib 			\masm32\lib\winmm.lib
includelib			\masm32\lib\masm32.lib
includelib			\masm32\lib\msvcrt.lib


include				cryptohash.inc			;for MD5 calculation
includelib			cryptohash.lib



;***********************************************************************
;* PROCS
;***********************************************************************
include				..\inc\hashtable.inc





;---private functions---
ht_get_item 			PROTO :DWORD,:DWORD,:DWORD
ht_get_item_by_index		PROTO :DWORD,:DWORD
ht_memorydump_calc_size		PROTO :DWORD
ht_item_get_key			PROTO :DWORD,:DWORD
ht_item_set_key			PROTO :DWORD,:DWORD
ht_convert_all_ht_to_dump	PROTO :DWORD


;***********************************************************************
;* STRUCTURES
;***********************************************************************
;---private structures---
MD5_CHECKSUM struct
	dword_1		DWORD ?
	dword_2		DWORD ?
	dword_3		DWORD ?
	dword_4		DWORD ?
MD5_CHECKSUM ends


HT_HEADER struct
	ht_head_sig_1	DWORD ? ; "HASH"
	ht_head_sig_2	DWORD ? ; "TABL"
	ht_heap		DWORD ?
	ht_items	DWORD ? ;pointer to array of pointers to HT_ITEM
	ht_itemcount	DWORD ?	;number of items
	ht_itemsize	DWORD ? ;how much memory itemdata need
	ht_compareto	DWORD ? ;pointer to compareTo functions, for sorting items
	ht_last_md5	MD5_CHECKSUM <>
HT_HEADER ends


HT_ITEM struct
	i_data		DWORD ? ;pointer to itemdata
	i_size		DWORD ? ;datasize
	i_key		MD5_CHECKSUM <>
	i_options	DWORD ?
HT_ITEM ends


;---memory dump structure---
;HT_HEADER
;array of HT_ITEM
;array of data from HT_ITEM



;***********************************************************************
;* CONSTANTS
;***********************************************************************
.const
;---options for ht_put method---
;-public-
ITEM_STRING_KEY		equ 1	;key is pointer to a string

;-private-
ITEM_DWORD		equ 2   ;data is a DWORD
ITEM_HASHTABLE		equ 4   ;data is a hashtable handle ;when its a memory dump data is a full dump of the nested hashtable
ITEM_MD5_KEY		equ 8   ;key is pointer to MD5_CHECKSUM structure
ITEM_DATA_POINTER	equ 16  ;data is stored in alloced memory


;***********************************************************************
;* DATA
;***********************************************************************
.data
sig_1			dd "HSAH"
sig_2			dd "LBAT"




;***********************************************************************
;* CODE
;***********************************************************************
.code

DllEntry proc _hinstance:DWORD, _reason:DWORD, _reserved1:DWORD
	mov eax,1 ;TRUE
	ret
DllEntry endp


ht_create proc public uses esi edi ebx
	
	fn HeapCreate,HEAP_NO_SERIALIZE,sizeof HT_HEADER,0
	mov ebx,eax
	
	
	fn HeapAlloc,ebx,HEAP_ZERO_MEMORY,sizeof HT_HEADER
	mov esi,eax
	mov [esi].HT_HEADER.ht_heap,ebx
	
	
	;---signature---
	m2m [esi].HT_HEADER.ht_head_sig_1,sig_1
	m2m [esi].HT_HEADER.ht_head_sig_2,sig_2
	
	
	;---memory for item pointers---
	fn HeapAlloc,ebx,HEAP_ZERO_MEMORY,DWORD
	mov [esi].HT_HEADER.ht_items,eax


	mov eax,esi	
	ret	;return pointer to HT_HEADER
ht_create endp


ht_close proc public uses esi edi ebx _ht

	LOCAl local_itemcount	:DWORD
	
	;---close---
	mov eax,_ht
	fn HeapDestroy,[eax].HT_HEADER.ht_heap
	
	ret
ht_close endp


ht_close_all proc public uses esi edi ebx _ht

	LOCAl local_itemcount	:DWORD
	
	;---parse nested hashtables and close them---
	mov local_itemcount,0
	.repeat
		fn ht_get_item_by_index,_ht,local_itemcount
		.if eax
			mov edi,eax
			
			.if [edi].HT_ITEM.i_options & ITEM_HASHTABLE
			
				fn ht_close_all,[edi].HT_ITEM.i_data
			.endif
			
			mov eax,TRUE
		.endif
		
		inc local_itemcount
	.until eax==0

	;---close---
	mov eax,_ht
	fn HeapDestroy,[eax].HT_HEADER.ht_heap
	
	ret
ht_close_all endp




ht_size proc public _ht
	;--get number of items---
	mov eax,_ht
	mov eax,[eax].HT_HEADER.ht_itemcount
	ret
ht_size endp


ht_put proc public uses esi edi ebx _ht,_key,_data,_datasize,_options
	
	LOCAL local_key		:DWORD
	
	mov esi,_ht
	
	
	;---does key already exist?---
	fn ht_get_item,_ht,_key,_options
	.if eax==0
		;---create new HT_ITEM object---
		fn HeapAlloc,[esi].HT_HEADER.ht_heap,HEAP_ZERO_MEMORY,sizeof HT_ITEM
		mov ebx,eax
		
		
		;---set itemcound---
		inc [esi].HT_HEADER.ht_itemcount
		
				
		;---set i_key (dont run md5 calculation twice,use result from ht_get_item)---
		m2m [ebx].HT_ITEM.i_key.dword_1,[esi].HT_HEADER.ht_last_md5.dword_1
		m2m [ebx].HT_ITEM.i_key.dword_2,[esi].HT_HEADER.ht_last_md5.dword_2
		m2m [ebx].HT_ITEM.i_key.dword_3,[esi].HT_HEADER.ht_last_md5.dword_3
		m2m [ebx].HT_ITEM.i_key.dword_4,[esi].HT_HEADER.ht_last_md5.dword_4
		
		
		;---realloc item pointer heapsize---
		mov ecx,[esi].HT_HEADER.ht_itemcount
		mov eax,DWORD
		mul ecx
		fn HeapReAlloc,[esi].HT_HEADER.ht_heap,HEAP_ZERO_MEMORY,[esi].HT_HEADER.ht_items,eax
		mov [esi].HT_HEADER.ht_items,eax
		
		
		;---store pointer to new HT_ITEM---
		mov eax,[esi].HT_HEADER.ht_items
		mov ecx,[esi].HT_HEADER.ht_itemcount
		mov dword ptr[eax+DWORD*ecx-DWORD],ebx
		
		
	.else
		;---overwrite existing dc item---
		mov ebx,eax
		
		;---free old data block---
		.if !([ebx].HT_ITEM.i_options & ITEM_DWORD) ;prevent crash
			fn HeapFree,[esi].HT_HEADER.ht_heap,0,[ebx].HT_ITEM.i_data 
		.endif
		
		;---sub overall itemdatasize---
		mov ecx,[ebx].HT_ITEM.i_size
		sub [esi].HT_HEADER.ht_itemsize,ecx
	.endif
	
	
	;---options---
	mov eax,_options
	mov [ebx].HT_ITEM.i_options,eax
	
	
	;---copy data to own memory block---
	.if _options & ITEM_DWORD
		;---do not alloc extra memory to store a DWORD value---
		
		m2m [ebx].HT_ITEM.i_data,_data
		
		;---remove this bitflag---
		.if [ebx].HT_ITEM.i_options & ITEM_DATA_POINTER
			xor [ebx].HT_ITEM.i_options,ITEM_DATA_POINTER
		.endif	
	.else
		fn HeapAlloc,[esi].HT_HEADER.ht_heap,HEAP_ZERO_MEMORY,_datasize
		mov [ebx].HT_ITEM.i_data,eax
		fn RtlMoveMemory,[ebx].HT_ITEM.i_data,_data,_datasize ;dest,source,size
		
		or [ebx].HT_ITEM.i_options,ITEM_DATA_POINTER
	.endif
	
	;---item size---
	m2m [ebx].HT_ITEM.i_size,_datasize
	
	mov ecx,_datasize
	add [esi].HT_HEADER.ht_itemsize,ecx


	;---return pointer to HT_ITEM---
	mov eax,ebx
	ret
ht_put endp


ht_put_string proc public _ht,_key,_string,_options
	
	;---get string size---
	fn szLen,_string
	inc eax			;for terminating 0
	
	;---copy string to our hashtable---
	fn ht_put,_ht,_key,_string,eax,_options
	ret
ht_put_string endp


ht_put_dword proc public _ht,_key,_dword,_options
	
	
	mov ecx,_options
	or ecx,ITEM_DWORD
	
	fn ht_put,_ht,_key,_dword,DWORD,ecx
	ret
ht_put_dword endp


ht_put_hashtable proc public _ht,_key,_ht_to_put,_options

	mov ecx,_options
	or ecx,ITEM_HASHTABLE
	
	fn ht_put_dword,_ht,_key,_ht_to_put,ecx
	ret
ht_put_hashtable endp	


ht_get_item_by_index proc private uses esi edi ebx _ht,_index

	mov esi,_ht
	
	xor eax,eax
	
	mov ecx,_index
	.if ecx < [esi].HT_HEADER.ht_itemcount
		
		mov edx,[esi].HT_HEADER.ht_items
		
		mov eax,[edx+ecx*DWORD]	
	.endif
	
	ret
ht_get_item_by_index endp


ht_get_item proc private uses esi edi ebx _ht,_key,_options
	
	LOCAL local_counter	:DWORD
	LOCAL local_found_item	:DWORD
	
	
	
	mov esi,_ht

	mov local_found_item,0


	;---i_key---
	.if _options & ITEM_MD5_KEY
	
		mov eax,_key
		jmp @set_md5_key
	.else
		.if _options & ITEM_STRING_KEY
		
			fn IsBadReadPtr,esi,2
			.if eax==0
				;---calculate MD5 Key from String---
				fn MD5Init
				
				fn szLen,_key		
				fn MD5Update,_key,eax
				
				fn MD5Final
				
				
				@set_md5_key:
				
				m2m [esi].HT_HEADER.ht_last_md5.dword_1,[eax].MD5_CHECKSUM.dword_1
				m2m [esi].HT_HEADER.ht_last_md5.dword_2,[eax].MD5_CHECKSUM.dword_2
				m2m [esi].HT_HEADER.ht_last_md5.dword_3,[eax].MD5_CHECKSUM.dword_3
				m2m [esi].HT_HEADER.ht_last_md5.dword_4,[eax].MD5_CHECKSUM.dword_4
				
			.else
				jmp @return
			.endif	
		.else	
			;---dont use MD5 for simple 32Bit keys---
			m2m [esi].HT_HEADER.ht_last_md5.dword_1,_key
			mov [esi].HT_HEADER.ht_last_md5.dword_2,0
			mov [esi].HT_HEADER.ht_last_md5.dword_3,0
			mov [esi].HT_HEADER.ht_last_md5.dword_4,0
		.endif	
	.endif

	
	
	
	;----
	mov edx,[esi].HT_HEADER.ht_items
	

	m2m local_counter,[esi].HT_HEADER.ht_itemcount
	
	.while local_counter!=0
		mov ebx,[edx]
		
		
		mov ecx,[esi].HT_HEADER.ht_last_md5.dword_1
		.if [ebx].HT_ITEM.i_key.dword_1!=ecx
			jmp @next
		.endif
		
		mov ecx,[esi].HT_HEADER.ht_last_md5.dword_2
		.if [ebx].HT_ITEM.i_key.dword_2!=ecx
			jmp @next
		.endif
		
		mov ecx,[esi].HT_HEADER.ht_last_md5.dword_3
		.if [ebx].HT_ITEM.i_key.dword_3!=ecx
			jmp @next
		.endif
		
		mov ecx,[esi].HT_HEADER.ht_last_md5.dword_4
		.if [ebx].HT_ITEM.i_key.dword_4!=ecx
			jmp @next
		.endif
		

		mov local_found_item,ebx
		.break
		
		
		@next:
		add edx,DWORD
		dec local_counter
	.endw
	
	@return:
	mov eax,local_found_item
	ret	;return pointer to HT_ITEM
ht_get_item endp


ht_get_item_data proc public uses ebx  _ht,_key,_options

	fn ht_get_item,_ht,_key,_options
	.if eax
		mov ebx,eax
		mov eax,[ebx].HT_ITEM.i_data
		
	.endif
	ret
ht_get_item_data endp


ht_get_item_size proc public _ht,_key,_options

	fn ht_get_item,_ht,_key,_options
	.if eax
		mov eax,[eax].HT_ITEM.i_size
	.endif
	ret
ht_get_item_size endp


ht_memorydump_calc_size proc private uses esi edi ebx _ht
	
	LOCAL local_size	:DWORD
	
	;---size of HT_HEADER + HT_ITEM + data

	mov esi,_ht

	;---HT_HEADER---
	mov local_size,sizeof HT_HEADER
	
	
	;---HT_ITEM---
	mov eax,[esi].HT_HEADER.ht_itemcount
	mov ecx,sizeof HT_ITEM
	mul ecx
	add local_size,eax
	
	
	;---item data size---
	mov ecx,[esi].HT_HEADER.ht_itemsize
	add local_size,ecx


	mov eax,local_size
	ret
ht_memorydump_calc_size endp



ht_copy_item_data proc public uses ebx _ht,_key,_options,_dest

	fn ht_get_item,_ht,_key,_options
	.if eax
		mov ebx,eax
		
		.if [ebx].HT_ITEM.i_options & ITEM_DWORD
			mov edx,_dest
			m2m [edx],[ebx].HT_ITEM.i_data
		.else
			fn RtlMoveMemory,_dest,[ebx].HT_ITEM.i_data,[ebx].HT_ITEM.i_size
		.endif
		
		mov eax,[ebx].HT_ITEM.i_size	;return size
	.endif
	ret
ht_copy_item_data endp


ht_remove proc public uses esi edi ebx _ht,_key,_options

	LOCAL local_index		:DWORD
	LOCAL local_shift_counts	:DWORD
	
	fn ht_get_item,_ht,_key,_options
	.if eax
		mov ebx,eax
		mov esi,_ht
		
		;---itemsize---
		mov ecx,[ebx].HT_ITEM.i_size
		sub [esi].HT_HEADER.ht_itemsize,ecx
	
	
		;---free mem---
		fn HeapFree,[esi].HT_HEADER.ht_heap,0,ebx
		
		
		;---get position in item pointer array---
		mov local_index,0
		mov edx,[esi].HT_HEADER.ht_items
		.while [edx]!=ebx
			inc local_index
			add edx,DWORD
		.endw	
		
		
		;---push from bottom to top---
		m2m local_shift_counts,[esi].HT_HEADER.ht_itemcount
		mov eax,local_index
		inc eax
		sub local_shift_counts,eax
		
		
		
		.while local_shift_counts!=0
			m2m [edx],[edx+DWORD]
			add edx,DWORD
			dec local_shift_counts
		.endw
		
		
		;---realloc item list mem---
		dec [esi].HT_HEADER.ht_itemcount
		mov ecx,[esi].HT_HEADER.ht_itemcount
		mov eax,DWORD
		mul ecx
		fn HeapReAlloc,[esi].HT_HEADER.ht_heap,HEAP_ZERO_MEMORY,[esi].HT_HEADER.ht_items,eax
		mov [esi].HT_HEADER.ht_items,eax
		
		
	.endif
	
	ret
ht_remove endp



ht_write_to_file proc public _ht,_filename
	
	LOCAL local_memsize 	:DWORD
	LOCAL local_dump	:DWORD
	
	fn ht_memorydump_create,_ht,addr local_memsize
	mov local_dump,eax
	
	
	fn write_disk_file,_filename,local_dump,local_memsize
	push eax
	
	fn ht_memorydump_close,local_dump
	
	pop eax
	ret
ht_write_to_file endp


ht_read_from_file proc public _filename
	
	LOCAL local_hFile		:DWORD
	LOCAL local_hFileMapping	:DWORD
	LOCAL local_hViewOfFile 	:DWORD
	
	LOCAL local_retvalue		:DWORD
	

	mov local_retvalue,0

	fn CreateFile,_filename,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,0
	.if eax!=INVALID_HANDLE_VALUE
		mov local_hFile,eax
		
		fn CreateFileMapping,eax,0,PAGE_READONLY,0,0,0
		.if eax!=NULL
			mov local_hFileMapping,eax
			
			fn MapViewOfFile,eax,FILE_MAP_READ,0,0,0
			.if eax!=NULL
				mov local_hViewOfFile,eax

				fn ht_create_from_memorydump,local_hViewOfFile
				mov local_retvalue,eax
				
				fn UnmapViewOfFile,local_hViewOfFile
			.endif
			
			fn CloseHandle,local_hFileMapping
		.endif
		
		fn CloseHandle,local_hFile
	.endif
	
	mov eax,local_retvalue
	ret
ht_read_from_file endp



ht_memorydump_create proc public uses esi edi ebx _ht,_out_memsize

	LOCAL local_memsize		:DWORD
	LOCAl local_itemcount		:DWORD
	LOCAL local_memorydump		:DWORD
	LOCAL local_ht_clone		:DWORD
	LOCAL local_ht_fileoffsets	:DWORD
	LOCAL local_HT_ITEM_offset	:DWORD
	LOCAL local_md5_key		:MD5_CHECKSUM
	
	;---hashtable which contains fileoffsets of the items, for later correction---
	fn ht_create
	mov local_ht_fileoffsets,eax
	
	
	;---clone hashtable for expanding nested hashtables---
	fn ht_clone,_ht
	mov local_ht_clone,eax
	
	
	;---replace all nested hashtables by memorydump---
	fn ht_convert_all_ht_to_dump,local_ht_clone
	
	
	;---get size---
	fn ht_memorydump_calc_size,local_ht_clone
	mov local_memsize,eax
	
	
	;---write memsize to output parameter---
	mov eax,_out_memsize
	.if eax
		m2m [eax],local_memsize
	.endif
	
	
	;---create heap for memdump---
	fn HeapCreate,HEAP_NO_SERIALIZE,local_memsize,0
	mov ebx,eax
	
	
	fn HeapAlloc,ebx,HEAP_ZERO_MEMORY,local_memsize
	mov esi,eax
	mov local_memorydump,eax
	
	;---copy HT_HEADER---
	fn RtlMoveMemory,esi,local_ht_clone,sizeof HT_HEADER
	
	;---correct heap handle---
	mov [esi].HT_HEADER.ht_heap,ebx
	add esi,sizeof HT_HEADER
	mov local_HT_ITEM_offset,esi
	

	;---copy HT_ITEM structures---
	mov edi,local_ht_clone
	m2m local_itemcount,[edi].HT_HEADER.ht_itemcount
	
	mov edi,[edi].HT_HEADER.ht_items
	.while local_itemcount!=0
	
		fn RtlMoveMemory,esi,[edi],sizeof HT_ITEM
		
		add edi,DWORD
		add esi,sizeof HT_ITEM
		dec local_itemcount
	.endw
	
	
	;---copy data---
	mov local_itemcount,0
	.repeat
		fn ht_get_item_by_index,local_ht_clone,local_itemcount
		.if eax
			mov edi,eax
			
			fn ht_item_get_key,edi,addr local_md5_key
			
			mov ecx,[edi].HT_ITEM.i_data
			or ecx,ITEM_MD5_KEY
			
			fn ht_copy_item_data,local_ht_clone,addr local_md5_key,ecx,esi
			
			;fn RtlMoveMemory,esi,[edi].HT_ITEM.i_data,[edi].HT_ITEM.i_size
			
			;---convert memory offsets to fileoffset---
			mov ecx,esi
			sub ecx,local_memorydump
		
			fn ht_put_dword,local_ht_fileoffsets,local_itemcount,ecx,0
			
			;---next---		
			add esi,[edi].HT_ITEM.i_size
			
			mov eax,TRUE
		.endif
		
		inc local_itemcount
	.until eax==0
	
	
	
	;---correct file offsets in HT_ITEM---
	
	mov esi,local_HT_ITEM_offset
	mov local_itemcount,0
	.repeat
		fn ht_get_item_by_index,local_ht_fileoffsets,local_itemcount
		.if eax
			mov ebx,eax
			
			fn ht_item_get_key,ebx,addr local_md5_key
			mov ecx, [ebx].HT_ITEM.i_options
			or ecx,ITEM_MD5_KEY
			
			fn ht_get_item_data,local_ht_fileoffsets,addr local_md5_key,ecx
			
			
			mov [esi].HT_ITEM.i_data,eax	;write file offset to data
			
			;---next---
			add esi,sizeof HT_ITEM
			mov eax,TRUE
		.endif
		
		inc local_itemcount
	.until eax==0
	
	
	
	;---close---
	fn ht_close,local_ht_fileoffsets
	fn ht_close,local_ht_clone
	
	
	mov eax,local_memorydump
	ret
ht_memorydump_create endp


ht_convert_all_ht_to_dump proc private uses esi edi ebx _ht
	
	LOCAL local_itemcount	:DWORD
	LOCAl local_ht		:DWORD
	LOCAL local_memdump	:DWORD
	LOCAL local_memsize	:DWORD
	LOCAL local_md5_key	:MD5_CHECKSUM
	
	;---copy data---
	mov local_itemcount,0
	.repeat
		fn ht_get_item_by_index,_ht,local_itemcount
		.if eax
			mov ebx,eax
			
			;---convert hashtable pointer to dump---
			.if [ebx].HT_ITEM.i_options & ITEM_HASHTABLE
			
				
			
				fn ht_item_get_key,ebx,addr local_md5_key
				mov ecx,[ebx].HT_ITEM.i_options
				or ecx,ITEM_MD5_KEY
				
				fn ht_get_item_data,_ht,addr local_md5_key,ecx
				mov local_ht,eax
				
				fn ht_memorydump_create,local_ht,addr local_memsize
				mov local_memdump,eax
				
				;---remove ITEM_DWORD---
				mov ecx,[ebx].HT_ITEM.i_options
				or ecx,ITEM_DWORD
				xor ecx,ITEM_DWORD
				
				
				
				fn ht_put,_ht,addr local_md5_key,local_memdump,local_memsize,ecx
				
				fn ht_memorydump_close,local_memdump
				
			.endif
			
			mov eax,TRUE
		.endif
		
		inc local_itemcount
	.until eax==0
	
	ret
ht_convert_all_ht_to_dump endp


ht_memorydump_close proc public _memdump

	mov eax,_memdump
	fn HeapDestroy,[eax].HT_HEADER.ht_heap

	ret
ht_memorydump_close endp


ht_create_from_memorydump proc public uses esi edi ebx _memdump

	LOCAL local_itemcount	:DWORD
	LOCAL local_ht		:DWORD
	LOCAL local_md5_key	:MD5_CHECKSUM
	
	
	mov esi,_memdump
	
	fn IsBadReadPtr,esi,8
	.if eax==0
	
		;---check for hashtable signature---
		mov eax,sig_1
		mov ecx,sig_2
		.if (([esi].HT_HEADER.ht_head_sig_1==eax) && ([esi].HT_HEADER.ht_head_sig_2==ecx))
			
			fn ht_create
			mov local_ht,eax
	
			m2m local_itemcount,[esi].HT_HEADER.ht_itemcount
			
	
			;---go to first HT_ITEM---
			add esi,sizeof HT_HEADER
						
						
			;---esi: points to first HT_ITEM---
			.while local_itemcount!=0
				
				fn ht_item_get_key,esi,addr local_md5_key
				
				;---convert fileoffset of data to memoryaddress--
				mov edx,[esi].HT_ITEM.i_data
				add edx,_memdump
				
				mov ecx,[esi].HT_ITEM.i_options
				or ecx,ITEM_MD5_KEY
				
				
				.if [esi].HT_ITEM.i_options & ITEM_HASHTABLE
					
					push ecx
					fn ht_create_from_memorydump,edx
					pop ecx	;new options with ITEM_MD5_KEY
					
					fn ht_put_hashtable,local_ht,addr local_md5_key,eax,ecx
				.else
					fn ht_put,local_ht,addr local_md5_key,edx,[esi].HT_ITEM.i_size,ecx
				.endif
				
				;---next---
				add esi,sizeof HT_ITEM
				dec local_itemcount
			.endw
			
			;---return handle to hashtable---
			mov eax,local_ht
		.else
			xor eax,eax
		.endif
	.else
		xor eax,eax
	.endif		
	ret
ht_create_from_memorydump endp


ht_clone proc public uses esi edi ebx _ht
	
	LOCAL local_itemcount	:DWORD
	LOCAL local_md5_key	:MD5_CHECKSUM
	
	fn ht_create
	mov edi,eax
	
	;---copy all items from source hashtable to new one---
	mov local_itemcount,0
	.repeat
		fn ht_get_item_by_index,_ht,local_itemcount
		.if eax
			mov ebx,eax
			
			fn ht_item_get_key,ebx,addr local_md5_key
			
			mov ecx,[ebx].HT_ITEM.i_options
			or ecx,ITEM_MD5_KEY
			
			fn ht_put,edi,addr local_md5_key,[ebx].HT_ITEM.i_data,[ebx].HT_ITEM.i_size,ecx
			
			fn ht_item_set_key,ebx,addr local_md5_key
			
			mov eax,TRUE
		.endif
		
		inc local_itemcount
	.until eax==0
	
	
	mov eax,edi
	ret
ht_clone endp




ht_item_get_key proc private uses esi edi ebx _ht_item,_out_key

	mov ebx,_ht_item
	mov edi,_out_key
	
	m2m [edi].MD5_CHECKSUM.dword_1,[ebx].HT_ITEM.i_key.dword_1
	m2m [edi].MD5_CHECKSUM.dword_2,[ebx].HT_ITEM.i_key.dword_2
	m2m [edi].MD5_CHECKSUM.dword_3,[ebx].HT_ITEM.i_key.dword_3
	m2m [edi].MD5_CHECKSUM.dword_4,[ebx].HT_ITEM.i_key.dword_4
	
	ret
ht_item_get_key endp


ht_item_set_key proc private uses esi edi ebx _ht_item,_key

	mov ebx,_ht_item
	mov edi,_key
	
	m2m [ebx].HT_ITEM.i_key.dword_1,[edi].MD5_CHECKSUM.dword_1
	m2m [ebx].HT_ITEM.i_key.dword_2,[edi].MD5_CHECKSUM.dword_2
	m2m [ebx].HT_ITEM.i_key.dword_3,[edi].MD5_CHECKSUM.dword_3
	m2m [ebx].HT_ITEM.i_key.dword_4,[edi].MD5_CHECKSUM.dword_4

	ret
ht_item_set_key endp

end DllEntry